#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/major.h>

#include <asm/uaccess.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/fcntl.h>
#include <linux/ioctl.h>
#include <asm/atomic.h>
#include "chr_drv.h"

MODULE_LICENSE("GPL");

//#define LOG_FUNC_CALL()		{printk("%s\n", __func__);}
#define LOG_FUNC_CALL1()	
#define LOG_FUNC_CALL()				

#define SUCCESS 0
#define CRIT_ERR -1

//void displayMemory(unsigned char *data); 
static int char_open(struct inode *inode, struct file *filp);
static ssize_t char_read(struct file *filp, char *buf, size_t count, loff_t *f_pos);
static ssize_t char_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos);
static int char_ioctl(struct inode *inode, struct file *filp, unsigned int ioctl_cmd, unsigned int ioctl_arg);
static int char_release(struct inode *inode, struct file *filp);
static unsigned int char_poll(struct file *filp, poll_table *wait);
static void close_disk(void);

void copy_from_disk(void *dst, sector_t sector, size_t len);
void copy_to_disk(const void *src, sector_t sector, size_t len);
int is_disk_server_enabled(void);

// Aliasing write, read, ioctl, etc...
static struct file_operations drv_interface = {
	.read =   char_read,
	.write =  char_write,
	.poll = char_poll,
	.ioctl =  char_ioctl,
	.open =   char_open,
	.release = char_release,
};

DECLARE_WAIT_QUEUE_HEAD(disk_wait);
DECLARE_WAIT_QUEUE_HEAD(blk_drv_sleep);
	
static int enable;


atomic_t cmd_status;
blk_drv_if  disk_server_if; 

static int __init char_init(void)
{
	int rc;
	LOG_FUNC_CALL();

	rc = register_chrdev(NI_CHAR_DRV_MAJOR_NUM, NI_CHAR_DRV_NAME, &drv_interface);
	if (rc < 0) {
  	printk(KERN_ERR "%s was not register\n",NI_CHAR_DRV_NAME);
  	return rc;
	}
	atomic_set(&cmd_status, 0);

	printk(KERN_INFO "char_drv: module loaded\n");
	return SUCCESS; 
}

static void __exit char_close(void)
{
	LOG_FUNC_CALL();
	unregister_chrdev(NI_CHAR_DRV_MAJOR_NUM, NI_CHAR_DRV_NAME);
}

module_init(char_init);
module_exit(char_close);


//----------------------------------------------------------------------------
// Char Driver Functions
//----------------------------------------------------------------------------
static int char_open(struct inode *inode, struct file *filp)
{
 	LOG_FUNC_CALL();
	/*Only when someone aka disk_server opens the port is when things are ready for block drivers*/
	enable = 1;
	return SUCCESS;
}

ssize_t char_read(struct file *filp, char *buf, size_t count, loff_t *f_pos)
{
	char *ptr; 
 	LOG_FUNC_CALL1();

	ptr = (char*)&disk_server_if; 
	if (copy_to_user(buf, ptr, count))  {
	  printk(KERN_INFO "Failed copy to user.\n");
  }
 	return (0);
}

ssize_t char_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos)
{
	char *ptr; 
 	LOG_FUNC_CALL1();	

	ptr = (char*) &disk_server_if;
	if (copy_from_user(ptr, buf, count)) {
    printk(KERN_INFO "Failed copy from user.\n");
  }
  return (0);
}

static int char_ioctl(struct inode *inode, struct file *filp, unsigned int ioctl_cmd, unsigned int ioctl_arg)
{
	int rc = SUCCESS;
 	LOG_FUNC_CALL();	

	switch (ioctl_cmd) {
	case FD_DONE: 
		atomic_set(&cmd_status, 0);
		__wake_up_sync(&blk_drv_sleep, TASK_NORMAL, 1);
	break;
	
	case FD_CLOSE: 
		close_disk();
	break; 

	default: 
		printk(KERN_INFO "No such ioctl");	
	}

	return rc;
}

static unsigned int char_poll(struct file *filp, poll_table *wait)
{
  unsigned int mask = 0;
 	LOG_FUNC_CALL1();	

	poll_wait(filp, &disk_wait, wait);
	
	if (atomic_read(&cmd_status) == 1)
		mask |= (POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM);   /* Readable or Writable */

	return mask;
}


static int char_release(struct inode *inode, struct file *filp)
{
 	LOG_FUNC_CALL();	
 	return SUCCESS;
}

//----------------------------------------------------------------------------
// Blk & Char Driver Interface 
//----------------------------------------------------------------------------
void copy_from_disk(void *dst, sector_t sector, size_t len)
{
 	LOG_FUNC_CALL();	
	disk_server_if.command = LBA_READ; //Read Cmd to disk_server 
	disk_server_if.blk_num = sector;
	disk_server_if.data_length = len;
	atomic_set(&cmd_status, 1);
}

void copy_to_disk(const void *src, sector_t sector, size_t len)
{
	LOG_FUNC_CALL();

	disk_server_if.command = LBA_WRITE;
	disk_server_if.blk_num = sector;
	disk_server_if.data_length = len;
	memcpy(disk_server_if.data, src, len);
	atomic_set(&cmd_status, 1);
}

static void close_disk(void)
{
	LOG_FUNC_CALL();

	disk_server_if.command = LBA_EXIT; //send lba_exit to disk_server 
	disk_server_if.blk_num = 0;
	disk_server_if.data_length = 0;

	atomic_set(&cmd_status,1);
	__wake_up_sync(&disk_wait, TASK_NORMAL, 1);
}

int is_disk_server_enabled(void)
{
 	LOG_FUNC_CALL();	
		return (enable); 
}

EXPORT_SYMBOL(disk_server_if);
EXPORT_SYMBOL(cmd_status);
EXPORT_SYMBOL(disk_wait);
EXPORT_SYMBOL(blk_drv_sleep);

EXPORT_SYMBOL(is_disk_server_enabled);
EXPORT_SYMBOL(copy_from_disk);
EXPORT_SYMBOL(copy_to_disk);



